'''
Since some elsevier open papers could not be processed with papers class (in papers/raw.py), many methods are overwritten in
paper_elsevier and papers_elsevier.

[Documentation here can be more detailed if needed]
'''

import os
import json
from tqdm.notebook import tqdm
from collections import Counter
import re
import pandas as pd
import ast

from naimai.utils.general import get_soup
from naimai.constants.fields import fields_codes_elsevier
from naimai.constants.paths import codes_fields_path, path_open_citations
from naimai.papers.raw import paper_base,papers
from naimai.utils.regex import multiple_replace
from naimai.decorators import update_naimai_dois

class elsevier_data:
  def __init__(self,myfield_dict):
    self.files_path = 'json/json/'
    self.all_files = os.listdir(self.files_path)
    self.files_codes = {}
    self.files_fields = {}
    self.myfield = myfield_dict

  def file_nb2data(self,nb):
    '''
    convert file_nb to content
    :param nb: nb of file (int)
    :return:
    '''
    file_path = os.path.join(self.files_path,self.all_files[nb])
    with open(file_path) as jf:
      data = json.load(jf)
    return data

  def file_nb2code(self,file_nb):
    '''
    convert file_nb to its elsevier codes
    :param file_nb: nb of file (int)
    :return:
    '''
    file_path = os.path.join(self.files_path,self.all_files[file_nb])
    with open(file_path) as jf:
      data = json.load(jf)
    return data['metadata']['asjc']

  def get_files_codes(self):
    '''
    for each file, get its elsevier codes
    :return:
    '''
    for nb in tqdm(range(len(self.all_files))):
      self.files_codes[nb]=self.file_nb2code(nb)

  def is_value_in_codes(self, list_codes, value):
    '''
    verify if code in list of codes (lst)
    :param list_codes: list of codes (list)
    :param value: code (str)
    :return:
    '''
    value = int(value)
    lst_split=[elt.split('-') for elt in list_codes]
    for elt in lst_split:
      if len(elt)>1:
        min_,max_ = int(elt[0]), int(elt[1])
        if value in range(min_, max_ + 1):
          return True
      else:
        if int(elt[0])==value:
          return True
    return False

  def codes2field(self, codes_lst):
    '''
    convert list of codes to the correct field of my fields
    :param codes_lst: list of codes (list)
    :return:
    '''
    result = []
    for key,val in self.myfield.items():
      for code in codes_lst:
        if self.is_value_in_codes(val, code):
          result.append(key)
          break
    return result


  def get_files_fields(self):
    '''
    get the correct field for all the files
    :return:
    '''
    all_files_len = len(self.all_files)
    for file_nb in tqdm(range(all_files_len)):
      code = self.file_nb2code(file_nb)
      self.files_fields[file_nb]= self.codes2field(code)

  def get_files_distribution(self):
    '''
    get the files distribution over my fields
    :return:
    '''
    if self.files_fields:
      result = []
      for val in self.files_fields.values():
        result+=val
      return Counter(result)
    return


class paper_elsevier(paper_base):
    def __init__(self, file_nb,json_data):
        super().__init__()
        self.file_name=file_nb
        self.json_data = json_data
        self.database = 'elsevier_kaggle'
        self.highlights_in = None
        self.doi = ''
        self.highlights = []

    def replace_abbreviations(self):
        abbreviations_dict = self.get_abbreviations_dict()
        if abbreviations_dict:
            self.Abstract = multiple_replace(abbreviations_dict, self.Abstract)
            self.Introduction = multiple_replace(abbreviations_dict, self.Introduction)
            self.Title = multiple_replace(abbreviations_dict, self.Title)
            self.Keywords = multiple_replace(abbreviations_dict, self.Keywords)
            self.highlights = [multiple_replace(abbreviations_dict, elt) for elt in self.highlights]

    def get_Abstract(self):
        if 'abstract' in self.json_data.keys():
            self.Abstract = self.json_data['abstract']

    def get_numCitedBy(self):
        if self.doi:
            path = path_open_citations + self.doi
            soup = get_soup(path)
            soup_list = ast.literal_eval(soup.text)
            if isinstance(soup_list,list):
                self.numCitedBy = len(soup_list)

    def get_highlights(self):
        if 'author_highlights' in self.json_data.keys():
            self.highlights_in = True
            self.highlights += [elt['sentence'] for elt in self.json_data['author_highlights']]
        else:
            self.highlights_in = False

    def title_sentence_from_elt(self, elt):
        result = ' '
        if 'title' in elt.keys():
            result += elt['title'] + '; '

        if 'sentence' in elt.keys():
            result += elt['sentence']
        return result

    def introduction_from_elt(self,elt):
        if 'title' in elt.keys():
            if re.findall('introduction', elt['title'], re.I):
               return elt['sentence']
        return
    def get_Introduction(self):
        if 'body_text' in self.json_data.keys():
            for elt in self.json_data['body_text']:
                stc = self.introduction_from_elt(elt)
                if stc:
                    self.Introduction += stc + ' '


    def conclusion_from_elt(self,elt):
        if 'title' in elt.keys():
            if re.findall('conclusion', elt['title'], re.I):
               return elt['sentence']
        return
    def get_Conslusion(self):
        if 'body_text' in self.json_data.keys():
            for elt in self.json_data['body_text']:
                stc = self.conclusion_from_elt(elt)
                if stc:
                    self.Conclusion += stc + ' '

    def get_raw_text(self):
        if 'body_text' in self.json_data.keys():
            for elt in self.json_data['body_text']:
                self.raw_text += self.title_sentence_from_elt(elt)

    def get_Title(self):
        if 'title' in self.json_data['metadata'].keys():
            self.Title = self.json_data['metadata']['title']

    def code2subfields(self,code,df_codes):
        return list(df_codes[df_codes['Code'] == int(code)][['Field', 'Subject area']].values[0])

    def get_fields(self,field=''):
        result = [field,]
        codes_fields_df = pd.read_excel(codes_fields_path)
        paper_codes = self.json_data['metadata']['asjc']
        for code in paper_codes:
            result += self.code2subfields(code, codes_fields_df)
        self.fields = list(set(result))

    def get_Authors(self):
        if 'authors' in self.json_data['metadata'].keys():
            first, last = '', ''
            authors_data = self.json_data['metadata']['authors']
            for elt in authors_data:
                if elt['first']:
                    first = elt['first']
                if elt['last']:
                    last = elt['last']
                self.Authors += ' ' + first+ ' '+ last +','
            if self.Authors[-1]==',':
                self.Authors= self.Authors[:-1]
        self.Authors = self.Authors.strip()

    def get_kwords(self):
        if 'keywords' in self.json_data['metadata'].keys():
            self.Keywords = ', '.join(self.json_data['metadata']['keywords'])

    def get_journal(self):
        if 'doi' in self.json_data['metadata'].keys():
            self.Journal = self.json_data['metadata']['issn']

    def get_doi(self):
        if 'doi' in self.json_data['metadata'].keys():
            self.doi = self.json_data['metadata']['doi']

    def get_year(self):
        if 'pub_year' in self.json_data['metadata'].keys():
            self.year = self.json_data['metadata']['pub_year']


class papers_elsevier(papers):
    def __init__(self,elsevier_data_obj=None):
        super().__init__() # loading self.naimai_dois & other attributes
        self.fields_dict = fields_codes_elsevier
        self.elsevier_data_obj = elsevier_data_obj
        self.files = []


    def get_all_files(self):
      self.elsevier_data_obj = elsevier_data(myfield_dict=fields_codes_elsevier)
      self.elsevier_data_obj.get_files_fields()

    def get_fields_files(self, field,reset=True):
      if reset:
          self.files=[]
      for k,v in self.elsevier_data_obj.files_fields.items():
        if field in v:
          self.files.append(k)

    def add_papers(self,paper_nb,field):
        json_data=self.elsevier_data_obj.file_nb2data(paper_nb)
        new_paper = paper_elsevier(file_nb=paper_nb,json_data=json_data)
        new_paper.get_doi()
        if not new_paper.is_in_database(self.naimai_dois):
             new_paper.get_fields(field)
             new_paper.get_Abstract()
             new_paper.get_Title()
             new_paper.get_Authors()
             new_paper.get_journal()
             new_paper.get_kwords()
             new_paper.get_year()
             new_paper.get_highlights()
             new_paper.replace_abbreviations()
             new_paper.get_numCitedBy()
             self.elements[new_paper.doi] = new_paper.save_dict()
             self.naimai_dois.append(new_paper.doi)

        else:
            pass
            # print('DOI {} already exists..'.format(new_paper.doi))



    @update_naimai_dois
    def get_papers(self,field, reset=True,update_dois=False,idx_start=0,idx_finish=-1):
        if reset:
            self.elements={}
        if not self.elsevier_data_obj:
            print('>> Getting all papers')
            self.get_all_files()
        self.get_fields_files(field)
        fles = self.files[idx_start:idx_finish]
        for f in tqdm(fles):
            try:
                self.add_papers(paper_nb=f, field=field)
            except:
                print('problem in paper ', f)



from unittest import TestCase
from naimai.papers.full_text.pdf_paper import papers_pdf
import os

class papers_pdfTest(TestCase):
    '''
    Test case papers issn object, by testing if the correct files are read, the title, numcitedby, database and journal are correct for the first paper.
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.fnames = ['Bousmar1999.pdf', 'Chen2016.pdf']
        cls.title = 'Determination of Apparent Shear Stress and its Application in Compound Channels'
        cls.numcitedby = float(10)
        cls.journal = 'Procedia Engineering'
        cls.database = "pdf"

        pdfs_path = os.path.join('tests', 'papers', 'input_data')
        obj = papers_pdf(papers_path=pdfs_path)
        obj.get_papers()
        result = obj.elements
        cls.keys=list(result.keys())
        cls.first_paper = result[cls.keys[1]]

    def test_fnames(self):
        '''
        test if correct fnames are read
        :return:
        '''
        self.assertEqual(self.keys,self.fnames)

    def test_title(self):
        '''
        test title of the first paper
        :return:
        '''
        title_first_paper = self.first_paper['Title']
        self.assertEqual(title_first_paper, self.title)

    def test_numcitedby(self):
        '''
        test numcitedby of the first paper
        :return:
        '''
        numcitedby_first_paper = self.first_paper['numCitedBy']
        self.assertEqual(numcitedby_first_paper, self.numcitedby)

    def test_journal(self):
        '''
        test journal of the first paper
        :return:
        '''
        journal_first_paper = self.first_paper['Journal']
        self.assertEqual(journal_first_paper, self.journal)

    def test_database(self):
        '''
        test database of the first paper
        :return:
        '''
        database_first_paper = self.first_paper['database']
        self.assertEqual(database_first_paper, self.database)


import os

grobid_url = "https://cloud.science-miner.com/grobid"
path_nlp = 'drive/MyDrive/MyProject/paper2/deps/SpaCy_Ner_data'
path_objective_classifier = 'naimai4science/objective-classifier-v0'
path_bomr_classifier = 'naimai4science/bomr-classifier-v0'
path_main_pipelines = os.path.join('main_pipelines_dir')
path_author_classifier = os.path.join('paper2','models','papers_classification','authors classif','authors_classif_GP')
codes_fields_path = "drive/MyDrive/MyProject/data_pipeline/landing_zone/Elsevier_Open_Journals/ASJC.xlsx"
naimai_dois_path = "drive/MyDrive/MyProject/data_pipeline/naimai_zone/naimai_dois"
path_formatted='drive/MyDrive/MyProject/data_pipeline/naimai_zone/Formatted_data/'
path_dispatched = 'drive/MyDrive/MyProject/data_pipeline/naimai_zone/Dispatched_data/'
path_produced = 'drive/MyDrive/MyProject/data_pipeline/naimai_zone/Production_data/'
path_similarity_model = os.path.join('drive/MyDrive/MyProject/data_pipeline','search_model')

path_google_search='https://google.com/search?q='
path_errors_log = 'drive/MyDrive/MyProject/errors_log/'
path_model_saving= 'drive/MyDrive/MyProject/models/PDF classif/'
path_open_citations = 'https://opencitations.net/index/coci/api/v1/citations/'

aws_root_pdfs = 'https://naimabucket.s3.eu-west-3.amazonaws.com/PDFs/'
Mybucket = "naimabucket"
arxiv_pdfs_url = 'https://arxiv.org/pdf/'

training_data_path = 'drive/MyDrive/MyProject/main_pipelines/all/total_training_papers'
naimai_data_path = 'drive/MyDrive/MyProject/main_pipelines/all/total_naimai_papers'
doi_url = 'https://www.doi.org/'

path_detailed_data = "drive/MyDrive/MyProject/data/Intent_classif_data/detailed_data.csv"
path_detailed_data_0_500= "drive/MyDrive/MyProject/data/Intent_classif_data/detailed_data_0-500.csv"
path_detailed_data_500_1000= "drive/MyDrive/MyProject/data/Intent_classif_data/detailed_data_500-1000.csv"
path_detailed_data_total = "drive/MyDrive/MyProject/data/Intent_classif_data/detailed_data_total.csv"

path_ner_data = "drive/MyDrive/MyProject/data/Intent_classif_data/NER_data_df.csv"
path_ner_balanced_data = "drive/MyDrive/MyProject/data/Intent_classif_data/path_ner_balanced_data.csv"
path_ner_balanced_data2 = "drive/MyDrive/MyProject/data/Intent_classif_data/path_ner_balanced_data2.csv"
path_ner_data_reduced = "drive/MyDrive/MyProject/data/Intent_classif_data/NER_data_reduced.csv"
path_ner_data_500_1000= "drive/MyDrive/MyProject/data/Intent_classif_data/NER_data_df_500-1000.csv"
path_ner_data_0_500= "drive/MyDrive/MyProject/data/Intent_classif_data/NER_data_df_0-500.csv"
path_ner_data_total = "drive/MyDrive/MyProject/data/Intent_classif_data/NER_data_df_total.csv"


from setuptools import find_packages, setup

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


setup(
    name = 'naimai',
    version = '1.0.0.4',
    author=  'Yassine Kaddi',
    author_email = 'yassine@naimai.fr',
    description = 'Python library to help with scientific literature research',
    long_description = long_description,
    install_requires = requirements,
    long_description_content_type = 'text/markdown',
    maintainer = "Yassine Kaddi",
    maintainer_email = "yassine@naimai.fr",
    license = "CC BY-NC-SA",
    keywords = ['science', 'review','bibliography','python','nlp','machine-learning','information-extraction'],
    url =" https://github.com/yassinekdi/naimai",
    classifiers =[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent"
    ],
    packages = find_packages()
)



from PyPDF2 import PdfFileWriter, PdfFileReader
import re
from tqdm.notebook import tqdm
import os

class PDFSpliter:
    def __init__(self):
        self.pdf = None
        self.abstract_pages = []
        self.abstract_pages_filtered = []  # remove papers with only 1 pages
        self.list_ranges = []

    def read_pdf(self, pdf_path):
        self.pdf = PdfFileReader(open(pdf_path, "rb"))

    def get_abstract_pages(self,spliter_pattern):
        '''
        get page numbers having abstract
        '''
        for i in tqdm(range(self.pdf.numPages)):
            output = PdfFileWriter()
            output.addPage(self.pdf.getPage(i))

            page = output.pages[0]
            text = page.extract_text()

            if re.findall(spliter_pattern, text, re.I):
                self.abstract_pages.append(i + 1)

    def filter_abstract_pages(self):
        '''
        remove papers with only 1 pages
        '''

        for idx, page in enumerate(self.abstract_pages[:-1]):
            if page == self.abstract_pages[idx + 1] + 1:
                pass
            else:
                self.abstract_pages_filtered.append(page)
        self.abstract_pages_filtered.append(page + 1)
        print(f'len abstract pages : {len(self.abstract_pages)} -- len filtered : {len(self.abstract_pages_filtered)}')

    def get_ranges(self):
        '''
        range of pages of each article
        '''
        for idx, elt in enumerate(self.abstract_pages_filtered[:-1]):
            self.list_ranges.append((elt, self.abstract_pages_filtered[idx + 1] - 1))

    def output_range(self, range_pages: tuple, file_name: str, output_dir=''):
        '''
        output range of pages in pdf doc
        '''
        p1, p2 = range_pages
        output = PdfFileWriter()

        for page in range(p1 - 1, p2):
            page = self.pdf.getPage(page)
            text = page.extract_text()

            if len(text.split()) > 20:
                output.addPage(page)

        output_path = os.path.join(output_dir, f"{file_name}_{p1}_{p2 + 1}.pdf")
        with open(output_path, "wb") as outputStream:
            output.write(outputStream)

    def split(self, pdf_path, output_dir='',spliter_pattern='abstract:|abstract\n'):
        '''
        split pdf_path
        '''
        self.read_pdf(pdf_path)
        print('>> Getting pdf pages..')
        self.get_abstract_pages(spliter_pattern)

        print('>> Filtering..')
        self.filter_abstract_pages()

        print('>> Get ranges..')
        self.get_ranges()

        print('>> Exporting..')
        file_name = pdf_path.split('/')[-1].replace('.pdf','')

        for range_pages in tqdm(self.list_ranges):
            self.output_range(range_pages, file_name, output_dir)


import os
from tqdm.notebook import tqdm
from naimai.constants.paths import path_dispatched, path_formatted, path_produced
from naimai.utils.general import load_gzip
from naimai.utils.regex import get_ref_url
import shutil
import matplotlib.pyplot as plt
import random
import re
from spacy_langdetect import LanguageDetector
from spacy.language import Language
import spacy
from naimai.constants.nlp import nlp_vocab
random.seed(10)

def create_lang_detector(nlp, name):
    return LanguageDetector()

class Zone:
    def __init__(self, zone_path, zone_name):
        self.elements = {}
        self.zone_path = zone_path
        self.zone_name = zone_name

    def get_papers(self, field, fname, verbose=True):
        '''
        load database
        :param database:
        :return:
        '''
        path_db = os.path.join(self.zone_path,field,fname)
        data=load_gzip(path_db)
        if verbose:
            print('Len data: ', len(data))
        return data

    def get_elements(self):
        if self.zone_path:
            elements = os.listdir(self.zone_path)
            for elt in elements:
                path_elt = os.path.join(self.zone_path, elt)
                if os.path.isdir(path_elt):
                    self.elements[elt] = [elt for elt in os.listdir(path_elt) if os.path.isfile(os.path.join(path_elt,elt)) and 'encoding' not in elt and 'sqlite' not in elt]
                else:
                    self.elements[elt] = ''

    def merge_elements(self,databases_names: list) -> dict:
        '''
        merge list of databases names
        :param elements:
        :return:
        '''
        merged = {}
        paths = [os.path.join(self.zone_path,db) for db in databases_names]

        for path in tqdm(paths):
            papers = load_gzip(path)
            merged.update(papers)
        return merged

    def create_fields(self, fields_names):
        for field in fields_names:
            path = os.path.join(self.zone_path, field)
            os.mkdir(path)

    def reset_all(self):
        '''
        clear the zone
        :return:
        '''
        elements = os.listdir(self.zone_path)
        for elt in elements:
            path = os.path.join(self.zone_path, elt)
            if os.path.isdir(path):
                os.rmdir(path)
            else:
                os.remove(path)

    def reset_all_elements(self):
        '''
        clear all elements (folders) of the zone
        :return:
        '''
        for element in self.elements:
            self.reset_element(element)

    def reset_element(self, element):
        '''
        clear the elements of the input element (field/folder)
        :param element: field/folder of the zone
        :return:
        '''
        if element in self.elements:
            path_element = os.path.join(self.zone_path, element)
            if os.path.isdir(path_element):
                files = os.listdir(path_element)
                for f in files:
                    path = os.path.join(path_element, f)
                    if os.path.isfile(path):
                        os.remove(path)
                    elif os.path.isdir(path):
                        shutil.rmtree(path)
            elif os.path.isfile(path_element):
                os.remove(path_element)
        else:
            print('{} is not an element in {} zone'.format(element, self.zone_name))


class Dispatched_Zone(Zone):
    def __init__(self):
        super().__init__(zone_path=path_dispatched, zone_name='dispatched')
        self.get_elements()

    def get_field(self,field,size_data=None,verbose=True):
        '''
        load database, if nb samples given, load nb samples of the database
        :param database:
        :return:
        '''
        size_each_all_papers = 0
        path_field = os.path.join(self.zone_path,field)
        all_papers = os.listdir(path_field)

        if size_data:
            size_each_all_papers = int(size_data/len(all_papers))+1
            print('>> Each all_papers size : ', size_each_all_papers)
        paths_papers = [os.path.join(path_field,pap) for pap in all_papers]
        data_list=[]
        
        for path_db,paper in zip(paths_papers, all_papers):
            data=load_gzip(path_db)
            if size_each_all_papers:
                keys = list(data.keys())
                if len(data)>size_each_all_papers:
                    keys_selected= random.sample(keys,size_each_all_papers)
                else:
                    keys_selected = keys
                data = {key: data[key] for key in keys_selected}
            data_list.append(data)
            if verbose:
                print(f'paper {paper} - Len data: {len(data)}')
        return data_list

    def reset_allpapers_elt(self,all_papers):
        '''
        reset 'all_papers' in all fields
        :param all_papers:
        :return:
        '''
        fields = list(self.elements)
        for field in fields:
            path = os.path.join(self.zone_path,field,all_papers)
            if os.path.isfile(path):
                os.remove(path)


    def plot_distribution(self, verbose=False):
        fields = list(self.elements)
        lens = []
        for field in fields:
            data_list = self.get_field(field=field, verbose=verbose)
            len_field = sum([len(elt) for elt in data_list])
            lens.append(len_field)

        x,y = fields, lens
        plt.figure(figsize=(10, 8))
        plt.rcParams.update({'font.size': 17})
        plt.barh(x, y, alpha=.5)
        plt.xlabel('N° of papers', fontsize=20)
        plt.xticks(fontsize=16)
        plt.yticks(fontsize=16)
        plt.ylabel('Field', fontsize=20)
        for ind, val in enumerate(y):
            plt.text(val + 3, ind - .25, str(val))

class Formatted_Zone(Zone):
    def __init__(self):
        super().__init__(zone_path=path_formatted, zone_name='formatted')
        self.get_elements()

    def get_database(self,database):
        '''
        load database
        :param database:
        :return:
        '''
        path_db = os.path.join(self.zone_path,database)
        data=load_gzip(path_db)
        print('Len data: ', len(data))
        return data

class Production_Zone(Zone):
    def __init__(self):
        super().__init__(zone_path=path_produced, zone_name='production')
        self.get_elements()

    def clean_papers(self,field: str,fname: str):
        '''
        clean already computed papers
        :param field:
        :param fname:
        :return:
        '''
        print('>> Getting papers')
        papers = self.get_papers(field,fname)
        print('>> Removing empty elements')
        new_papers = self.remove_empty_elts(papers)
        print('>> Removing nonenglish papers')
        new_papers = self.remove_nonenglish_elts(new_papers)
        print('>> Correcting years')
        new_papers = self.correct_years(new_papers)
        print('>> Adding numCited')
        new_papers = self.add_numCitedBy(field,fname,new_papers)
        print('>> Adding authors')
        new_papers = self.add_allauthors(field,fname,new_papers,verbose=False)
        print('Pmc websites are not taken here!')
        print('>> New length : ', len(new_papers))
        return new_papers

    def remove_nonenglish_elts(self,papers: dict):
        '''
        remove non english papers from dictionary papers
        :param papers:
        :return:
        '''
        nlp = spacy.load(nlp_vocab)
        Language.factory("language_detector", func=create_lang_detector)
        nlp.add_pipe('language_detector', last=True)
        cleaned_paps = {}

        for fname in tqdm(papers):
            paper = papers[fname]
            if paper['messages']:
                text = nlp(paper['messages'][0])
            else:
                text = nlp(paper['title'])

            language_score=text._.language
            condition_english = (language_score['language'] == 'en') and (language_score['score'] > 0.7)
            if condition_english:
                cleaned_paps[fname] = paper

        return cleaned_paps

    def get_omr_dicts(self,fname: str,papers: dict) -> dict:
        '''
        for an fname, get obj, methods & results dict
        :param fname:
        :return:
        '''
        paper_name = '_'.join(fname.split('_')[:-1])
        omr_fnames = [paper_name + '_objectives', paper_name + '_methods', paper_name + '_results']

        result = {}
        for fname in omr_fnames:
            if fname in papers:
                result[fname] = papers[fname]
        return result

    def remove_empty_elts(self,papers: dict):
        '''
        remove empty dictionaries from all_papers dict & obj papers that has no messages and no methods nor results elements.
        :param papers:
        :return:
        '''
        cleaned_paps = {}
        for fname in papers:
            paper = papers[fname]
            if paper['messages']:
                cleaned_paps[fname] = papers[fname]
                continue
            if '_objectives' in fname:
                paper_elts = self.get_omr_dicts(fname,papers)
                is_empty= all([paper_elts[elt]['messages']==[] for elt in paper_elts])
                if not is_empty:
                    cleaned_paps[fname] = papers[fname]
        return cleaned_paps

    def correct_years(self,papers):
        '''
        correct some years wrongly saved: in format of xxxx.0
        :param papers:
        :return:
        '''
        new_papers = papers.copy()
        for fname in new_papers:
            if '_objectives' in fname:
                try:
                    new_papers[fname]['year'] = int(new_papers[fname]['year'])
                    new_papers[fname]['authors'] = re.sub('(\d)\.\d', r'\g<1>', new_papers[fname]['authors'])
                except:
                    print('problem in ', fname)
                if new_papers[fname]['reported']:
                    new_papers[fname]['reported'] = re.sub('(\d)\.\d', r'\g<1>', new_papers[fname]['reported'])
        return new_papers

    def add_numCitedBy(self,field,papers_name: str, produced_papers) -> dict:
        '''
        add numCitedBy parameter in produced papers when forgotten
        :param field:
        :param papers:
        :return:
        '''
        # get dispatched papers

        disp_zone = Dispatched_Zone()
        dispatched_papers = disp_zone.get_papers(field,papers_name)

        # add numCitedBy
        for key in tqdm(dispatched_papers):
            key_produced = key+'_objectives'
            if key in produced_papers:
                if 'numCitedBy' not in produced_papers[key_produced]:
                    numCitedBy = dispatched_papers[key]['numCitedBy']
                    if key_produced in produced_papers:
                        produced_papers[key_produced]['numCitedBy'] = numCitedBy

        return produced_papers

    def add_allauthors(self,field: str, papers_name: str,produced_papers={},verbose=True) -> dict:
        '''
        add all authors in produced papers
        :param field:
        :param papers_name:
        :param produced_papers:
        :return:
        '''

        if verbose:
            print('>> Getting papers')
        disp_zone = Dispatched_Zone()
        dispatched_papers = disp_zone.get_papers(field, papers_name)
        if not produced_papers:
            produced_papers = self.get_papers(field, papers_name)

        if verbose:
            print('>> Getting authors')
        for key in tqdm(dispatched_papers):
            key_produced = key+'_objectives'
            if key_produced in produced_papers and 'allauthors' not in produced_papers[key_produced]:
                produced_papers[key_produced]['allauthors'] = dispatched_papers[key]['Authors']

        return produced_papers

    def add_years_numCitedBy_in_all_data(self,field: str, papers_name: str) -> dict:
        '''
        add years in all elements
        :param field:
        :param papers_name:
        :return:
        '''
        produced_papers = self.get_papers(field, papers_name)
        new_produced_papers = {}
        pbs=0
        pattern = '_objectives|_methods|_results'
        for fname in tqdm(produced_papers):
            new_produced_papers[fname] = produced_papers[fname]
            if fname.endswith('_objectives'):
                pass
            else:
                root_fname = re.sub(pattern,'',fname) + '_objectives'
                try:
                    if not 'year' in new_produced_papers[fname]:
                        new_produced_papers[fname]['year']=produced_papers[root_fname]['year']
                        new_produced_papers[fname]['numCitedBy'] = produced_papers[root_fname]['numCitedBy']
                except:
                    pbs+=1
        print('>> Pbs in : ', pbs)
        return new_produced_papers

    def correct_pmc_websites(self,papers):
        '''
        some pmc papers websites were not correctly considered..
        :param papers:
        :return:
        '''
        for key in papers:
            if '_objectives' in key:
                doi = key.replace('_objectives','')
                papers[key]['website'] = get_ref_url(papers[key],doi)
        return papers



from bs4 import BeautifulSoup
from naimai.crawlers.issn_crawler import ISSN_crawler
from tqdm.notebook import tqdm
import random
import time

class doij_crawler:
  def __init__(self,html):
    self.soup = BeautifulSoup(html,features="lxml")
    self.cards = []
    self.docs = {'title': [], 'authors': [], 'date': [], 'fields': [], "abstract": [], "doi": [],
                     "numCitedBy": [], 'journals': []}

    self.get_cards()

  def get_cards(self):
    self.cards=self.soup.find_all(name="li", attrs={"class": "card search-results__record"})

  def get_journal(self,card,len_dois):
    heading = card.find(name="h3", attrs={"class": "search-results__heading"})
    return [heading.text.strip()]*len_dois

  def get_issn(self,card):
    heading = card.find(name="h3", attrs={"class": "search-results__heading"})
    issn = heading.find(name="a")['href'].split('toc')[1][1:]
    return issn

  def get_fields(self,card,len_dois):
    listing = card.find(name='div', attrs={'class': 'search-results__body'})
    listing_lists = listing.find_all(name='ul')
    return [[elt.text for elt in listing_lists[1].find_all(name='li')]]*len_dois

  def get_data_with_issn(self,issn, show_issn_tqdm=False,i1=0,i2=-1,verbose_every=0):
    cw = ISSN_crawler(issn=issn,field_issn='')
    try:
      cw.get_docs(idx_start=i1,idx_finish=i2,show_tqdm=show_issn_tqdm,verbose_every=verbose_every)
      return cw.docs
    except:
      return

  def get_dois(self,idx_start=0, idx_end=-1,t_min=2, t_max=5):
      cards = self.cards[idx_start:idx_end]
      for idx,card in tqdm(enumerate(cards), total=len(cards)):
          issn = self.get_issn(card)
          cw = ISSN_crawler(issn=issn, field_issn='')
          try:
              print('idx: ', idx)
              cw.get_dois(idx_start=0,idx_finish=-1)
          except:
              print('problem in idx : ', idx)
          self.docs['doi']+=cw.docs['doi']
          slp = random.randint(t_min, t_max)
          time.sleep(slp)

  def get_docs(self, idx_start=0, idx_end=-1,i1=0,i2=-1,show_issn_tqdm=False,verbose_every=0):
      for idx,card in tqdm(enumerate(self.cards[idx_start:idx_end]),total=len(self.cards[idx_start:idx_end])):
          issn = self.get_issn(card)
          issn_docs = self.get_data_with_issn(issn,show_issn_tqdm=show_issn_tqdm,i1=i1,i2=i2,verbose_every=verbose_every)
          if issn_docs:
            self.docs['abstract']+=issn_docs['abstract']
            if self.docs['abstract']:
              self.docs['title']+=issn_docs['title']
              self.docs['authors']+= issn_docs['authors']
              self.docs['date']+= issn_docs['date']
              self.docs['doi']+=issn_docs['doi']
              self.docs['numCitedBy']+=issn_docs['numCitedBy']

              self.docs['journals']+=self.get_journal(card, len(issn_docs['doi']))
              self.docs['fields']+=self.get_fields(card, len(issn_docs['doi']))




import re
from tqdm.notebook import tqdm
from bs4 import BeautifulSoup
import random
import requests
import time
from naimai.constants.regex import regex_paper_year


class HAL_crawler:
    def __init__(self, path, hal_field, t_min=1, t_max=4):
        # path= 'https://hal.archives-ouvertes.fr/search/index/?q=%2A&sort=submittedDate_tdate+desc&docType_s=ART&language_s=en&level0_domain_s=sde&page=XXX'
        self.path = path
        self.soup = {}
        self.hal_field = hal_field
        self.docs = {'title': [], 'authors': [], 'date': [], 'fields': [], "abstract": [], 'keywords': [], "doi": [],
                     'journal': [], 'hal_address': []}
        self.t_min = t_min
        self.t_max = t_max
        self.divs = []

    def get_soup(self, path, timeit):
        header = {}
        header[
            'User-Agent'] = 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36'
        soup = BeautifulSoup(requests.get(path, headers=header, timeout=60).content, 'html.parser')
        slp = random.randint(self.t_min, self.t_max)
        if timeit:
            time.sleep(slp)
        return soup

    def get_divs_page(self, soup_page):
        divs = soup_page.find_all('div')
        divs_filtered = [elt for elt in divs if elt.find_all(name='span', attrs={
            'class': "ref-authors"})]  # remove other divs than those of articles
        divs_filtered2 = [elt for elt in divs_filtered if
                          len(elt.find_all(name='strong')) == 1]  # remove some divs that gather all articles
        return divs_filtered2

    def get_dois_page_and_divs_filtered(self, divs_filtered):
        dois = []
        divs_filtered3 = []
        for elt in divs_filtered:
            doi = elt.find(name='a', attrs={'target': '_blank'})
            if doi:
                doi = doi.text.replace('⟨', '').replace('⟩', '')
                if doi not in dois:
                    dois.append(doi)
                    divs_filtered3.append(elt)
        return (dois, divs_filtered3)

    def get_titles_page(self, divs_filtered):
        titles = [elt.find('strong').text for elt in divs_filtered]
        return titles

    def get_authors(self,divs_filtered):
        authors = [elt.find(name='span', attrs={'class': "ref-authors"}).text.replace('\xa0',' ') for elt in divs_filtered]
        return authors

    def get_hal_addresses_page(self, divs_filtered):
        hal_addresses = [elt.find(name='a', attrs={'class': 'ref-halid'}).text for elt in divs_filtered]
        return hal_addresses

    def get_journals_page(self, soup_article):
        # journals = []
        div = soup_article.find('div', attrs={'class': "widget-content ref-biblio"})
        if div.find('i'):
            return div.find('i').text
        return ''
        # for elt in divs_filtered:
        #     for journal in elt.find_all('i'):
        #         if journal.text != 'et al.':
        #             journals.append(journal.text)
        #
        # return journals

    def get_years_pages(self, divs_filtered,dois):
        dates = []
        for idx, elt in enumerate(divs_filtered):
            doi = dois[idx]
            all_txt = elt.find_all('br')[1].text.replace(doi, '')

            year = re.findall(regex_paper_year, all_txt)
            if year:
                year = year[0]
            elif not year and 'In press' in all_txt:
                year = '2022'
            else:
                year = ''
            dates.append(year)
        return dates

    def get_soup_article(self, hal_address,doi):
        path_article = 'https://hal.archives-ouvertes.fr/' + hal_address
        try:
            soup_article = self.get_soup(path_article, timeit=True)
            return soup_article
        except:
            print('problem in doi : ', doi)
        return ''


    def get_abstract_article(self, soup_article):
        text = ''
        abstract = soup_article.find(name='div', attrs={'class': 'abstract-content'})
        if abstract:
            text = abstract.text
            text = text.replace('\nAbstract : ', '')
        return text

    def get_fields_article(self, soup_article):
        fields = self.hal_field
        field_elt = soup_article.find(name='blockquote', attrs={'style': 'margin-left:65px;margin-top:-15px;'})
        if field_elt:
            fields += ', ' + field_elt.find('strong').text.replace('/', ',').strip()
        return fields

    def get_keywords(self, soup_article):
        keywords_elt = soup_article.find(name='div', attrs={'class': 'keywords'})
        if keywords_elt:
            keywords = keywords_elt.text.replace('\n\nKeywords :\n', '').replace('\n\n', '').replace('\n', ', ').strip()
            return keywords
        return ''

    def get_nb_pages(self):
        path = self.path + '1'
        soup = self.get_soup(path, timeit=False)
        total_pages_soup = soup.find(name='ul', attrs={'class': 'pagination pagination-sm'}).find_all('li')[-1]
        total_pages = total_pages_soup.find('a')['href'].split('page=')[-1]
        return total_pages

    def get_soups(self,first_page,last_page):
      range_ = range(first_page, last_page + 1)
      for page in tqdm(range_, total=len(range_)):
            path = self.path + str(page)
            soup_page = self.get_soup(path, timeit=True)
            self.soup[page] = soup_page

    def get_docs(self):
        print('First data..')
        # soups = [BeautifulSoup(soup_df) for soup_df in soups_df['0']]
        for page in tqdm(self.soup):
            soup_page = self.soup[page]
            divs = self.get_divs_page(soup_page)
            dois, divs_filtered = self.get_dois_page_and_divs_filtered(divs)
            self.docs['doi']+= dois
            self.docs['title']+= self.get_titles_page(divs_filtered)
            self.docs['date']+= self.get_years_pages(divs_filtered,dois)
            self.docs['hal_address']+= self.get_hal_addresses_page(divs_filtered)
            self.docs['authors']+= self.get_authors(divs_filtered)

        print('Abstracts..')
        zip_ = zip(self.docs['doi'], self.docs['hal_address'])
        for doi, hal_address in tqdm(zip_, total=len(self.docs['doi'])):
            soup_article = self.get_soup_article(hal_address,doi)
            if soup_article:
                try:
                    abstract = self.get_abstract_article(soup_article)
                except:
                    abstract=''

                try:
                    fields = self.get_fields_article(soup_article)
                except:
                    fields= ''

                try:
                    keywords = self.get_keywords(soup_article)
                except:
                    keywords= ''

                try:
                    journal = self.get_journals_page(soup_article)
                except:
                    journal=''
            else:
                abstract,fields,keywords='','',''

            self.docs['abstract'].append(abstract)
            self.docs['fields'].append(fields)
            self.docs['keywords'].append(keywords)
            self.docs['journal'].append(journal)




'''
- Paper_Producer Class : takes a paper (as a dictionary) and structures its abstract into objectives, methods and results.

- Field_Producer Class : takes a field (directory with papers of the same field) and produces all the papers using Paper_Producer object.

- Custom_Producer Class : takes a pdf papers in dict format and structures them into abstract into objectives, methods and results.

Semantics using faiss & bert are commented
'''

from naimai.models.papers_classification.obj_classifier import Objective_classifier
from naimai.models.papers_classification.bomr_classif.ner_classifier import NER_BOMR_classifier
from naimai.models.papers_classification.bomr_classif.omr_segmentors import segment
from naimai.constants.nlp import nlp_vocab
from naimai.constants.regex import regex_not_converted2
from naimai.utils.regex import get_ref_url
from naimai.constants.paths import path_produced, path_dispatched, path_bomr_classifier
from naimai.utils.general import save_gzip, load_gzip, load_gzip_and_update
from naimai.models.text_generation.paper2reported import Paper2Reported
from naimai.pipelines.zones import Production_Zone
from naimai.models.papers_classification.semantic_search import Search_Model
import os
import re
import spacy
import random
from tqdm.notebook import tqdm
from sentence_transformers import SentenceTransformer
# import faiss
import numpy as np

# D = 768
# vecs = faiss.IndexFlatIP(D)
# ncentroids = 256
# code_size = 32


class Paper_Producer:
    '''
    Takes formatted paper 'fname' & transforms it to production paper, a dict of 3 differents dicts (obj,meth and res) :
    {'objectives': {'website': xx,
             'year':xx,
             'database': xx,
             'authors': et al,
             'messages': xx,
             'reported': xx,
             'title': xx,
             'journal': xx,
             ' numCitedBy': xx},
     'methods': {'messages': xx,
                'year': xx,
                ' numCitedBy': xx},
     'results': {'messages': xx,
                'year': xx,
                ' numCitedBy': xx}
     }
    When no objective is found after segmentation, we correct by looking for it in results, otherwise in methods.
    '''

    def __init__(self, paper: dict, paper_name='', obj_classifier=None, bomr_classifier=None, nlp=None,is_custom=False):
        self.paper_name = paper_name
        self.paper = paper
        self.authors = ''
        self.bomr_classifier = None
        self.obj_classifier = None
        self.is_custom= is_custom
        self.nlp = None
        self.production_paper = {}
        self.omr = {'objectives': [], 'methods': [], 'results': [], 'other': []}
        self.reported = ''
        self.smodel = None

        self.load_obj_classifier(obj_classifier)
        self.load_bomr_classifier(bomr_classifier)
        self.load_nlp(nlp)

    def load_obj_classifier(self, obj_classifier):
        '''
        Load the objective classifier: model that selects the objective sentence in the abstract.
        :param obj_classifier:
        :return:
        '''
        if obj_classifier:
            self.obj_classifier = obj_classifier
        else:
            print('>> Loading obj classifier..')
            self.obj_classifier = Objective_classifier()

    def load_bomr_classifier(self, bomr_classifier):
        '''
        Load the bomr classifier, the model that structures the abstract into bibliography, objectives, methods
        and results.
        :param bomr_classifier:
        :return:
        '''
        if bomr_classifier:
            self.bomr_classifier = bomr_classifier
        else:
            print('>> Loading bomr classifier..')
            self.bomr_classifier = NER_BOMR_classifier(load_model=True, path_model=path_bomr_classifier,
                                                       predict_mode=True)

    def clean_abstract(self,text: str) -> str:
        '''
        clean abstract by removing the non converted expressions
        :param text:
        :return:
        '''
        return re.sub(regex_not_converted2,'',text)

    def load_nlp(self, nlp):
        '''
        Load nlp model (spaCy)
        :param nlp:
        :return:
        '''
        if nlp:
            self.nlp = nlp
        else:
            print('>> Loading nlp..')
            self.nlp = spacy.load(nlp_vocab)

    def get_omr_dict(self):
        '''
        get {'obj': [xx,yy], 'methods': [xx,yy], 'results': [zz,ff]}
        :return:
        '''
        abstract = self.clean_abstract(self.paper['Abstract'])
        segments, sentences, _, _ = segment(text=abstract,
                                            obj_clf=self.obj_classifier,
                                            bomr_clf=self.bomr_classifier,
                                            summarize=True,
                                            check_bg=True,
                                            visualize_=False,
                                            nlp = self.nlp,
                                            )
        for stc_idx in segments:
            label = segments[stc_idx]
            self.omr[label].append(sentences[stc_idx])

    def report(self):
        '''
        report objectives (write objective sentences in reported speech), else report results, else methods
        :return:
        '''
        objectives, methods, results = self.omr['objectives'], self.omr['methods'], self.omr['results']
        if objectives:
            messages_reported = self.report_messages(objectives)
            self.reported = messages_reported
        if methods and not self.reported:
            messages_reported = self.report_messages(methods)
            self.reported = messages_reported
        if results and not self.reported:
            messages_reported = self.report_messages(results)
            self.reported = messages_reported

    def report_messages(self, messages: list) -> list:
        '''
        takes the messages and transform them into reported speech format.
        :param messages: list of sentences
        :return:
        '''
        messages_reported = []
        review = Paper2Reported(paper=self.paper,
                                paper_name=self.paper_name,
                                messages=messages,
                                nlp=self.nlp,
                                )
        review.generate()
        review.choose_objective()
        self.authors = review.processed_authors
        if review.list_reported:
            messages_reported = review.reported_objective
        return messages_reported

    def format_paper(self):
        '''
        Format the paper (in self.paper) and structures it into a dictionary with objectives, methods and results.
        :return:
        '''
        journal = self.paper['Journal']
        if journal and re.findall('[a-zA-Z]', journal):
            journal = re.sub('^nd|ISSN','',journal).strip()
        else:
            journal=''

        if self.paper['year']:
            year = int(self.paper['year'])
        else:
            year = self.paper['year']
            print('Year problem in paper: ', self.paper_name)

        if self.is_custom:
            objectives = {"website": '',
                          "year": year,
                          "database": self.paper['database'],
                          "messages": self.omr['objectives'],
                          "reported": self.reported,
                          "title": self.paper['Title'],
                          'numCitedBy': self.paper['numCitedBy'],
                          "journal": journal,
                          "authors": self.authors,
                          'allauthors': self.paper['Authors']}
        else:

            objectives = {"website": get_ref_url(self.paper),
                          "year": year,
                          "database": self.paper['database'],
                          "messages": self.omr['objectives'],
                          "reported": self.reported,
                          "title": self.paper['Title'],
                          'numCitedBy': self.paper['numCitedBy'],
                          "journal": journal,
                          "authors": self.authors,
                          'allauthors': self.paper['Authors']}
        methods = {"messages": self.omr['methods'],"numCitedBy": self.paper['numCitedBy'],"year": year}
        results = {"messages": self.omr['results'],"numCitedBy": self.paper['numCitedBy'],"year": year}
        self.production_paper = {'objectives': objectives, "methods": methods, "results": results}

    def produce_paper(self):
        '''
        Produce the paper by reporting its objectives sentence (in reported speech format) and formatting it
        (finding the objectives, methods and results)
        :return:
        '''
        self.get_omr_dict()
        objectives, methods, results = self.omr['objectives'], self.omr['methods'], self.omr['results']
        if objectives or methods or results:
            self.report()
            self.format_paper()


class Field_Producer:
    '''
    1 Takes formatted papers of a all_papers in a field (dispatched zone) and transform them into a produced all_paper using Paper Producer obj
    2 Can finetune search model to get the field encoder
    3 Compute field Faiss index
    '''
    def __init__(self, field, all_papers='', ref_fields= [], obj_classifier=None, bomr_classifier=None, nlp=None, encoder=None, field_papers=None, idx_start=0, idx_finish=-1, load_dispatched_field_papers=True, load_obj_classifier=True,
                 load_bomr_classifier=True, load_nlp=True, load_field_encoder=True):
        self.field = field
        self.dispatched_field_papers = {}
        self.obj_classifier = None
        self.bomr_classifier = None
        self.nlp = None
        self.encoder = None
        self.smodel = None
        self.ref_fields = ref_fields
        self.papers_ref_fields = {} # fields used to check if same paper is already produced

        self.produced_field_papers = {}
        self.field_index = None
        self.produce_only_fnames=False
        self.all_papers = all_papers
        if load_obj_classifier:
            self.load_obj_classifier(obj_classifier)
        if load_bomr_classifier:
            self.load_bomr_classifier(bomr_classifier)
        if load_nlp:
            self.load_nlp(nlp)
        if load_field_encoder:
            self.load_encoder(encoder)
        if load_dispatched_field_papers:
            self.load_dispatched_field_papers_(field_papers=field_papers, all_papers=all_papers, idx_start=idx_start, idx_finish=idx_finish)
        self.idx_finish = idx_finish
        self.idx_start= idx_start
        if idx_finish!=-1 or idx_start!=0:
            self.extracted=True
        else:
            self.extracted=False


    def load_obj_classifier(self, obj_classifier):
        '''
        Load the objective classifier: model that selects the objective sentence in the abstract.
        :param obj_classifier:
        :return:
        '''
        if obj_classifier:
            self.obj_classifier = obj_classifier
        else:
            print('>> Loading obj classifier..')
            self.obj_classifier = Objective_classifier()

    def load_bomr_classifier(self, bomr_classifier):
        '''
        Load the bomr classifier, the model that structures the abstract into bibliography, objectives, methods
        and results.
        :param bomr_classifier:
        :return:
        '''
        if bomr_classifier:
            self.bomr_classifier = bomr_classifier
        else:
            print('>> Loading bomr classifier..')
            self.bomr_classifier = NER_BOMR_classifier(load_model=True, path_model=path_bomr_classifier,
                                                       predict_mode=True)

    def load_nlp(self, nlp):
        '''
        Load nlp model (spaCy)
        :param nlp:
        :return:
        '''
        if nlp:
            self.nlp = nlp
        else:
            print('>> Loading nlp..')
            self.nlp = spacy.load(nlp_vocab)

    def load_encoder(self, encoder):
        '''
        Load the bert encoder, finetuned on the field produced papers.
        :param nlp:
        :return:
        '''
        if encoder:
          self.encoder = encoder
        else:
          path = os.path.join(path_produced, self.field, 'search_model')
          if os.path.isdir(path):
            print('>> Loading field encoder..')
            self.encoder = SentenceTransformer(path)
          else:
              print('>> No field encoder.. You need to fine tune !')

    def load_dispatched_field_papers_(self, field_papers: str, all_papers: str, idx_start=0, idx_finish=-1):
      '''
      Load field papers from dispatched zone
      :param field_papers: field of papers
      :param all_papers: all_papers name
      :param idx_start:
      :param idx_finish:
      :return:
      '''
      if field_papers:
          keys = list(field_papers)[idx_start:idx_finish]
          self.dispatched_field_papers = {elt: field_papers[elt] for elt in keys}
      else:
        print('>> Loading field papers : ', all_papers)
        path = os.path.join(path_dispatched,self.field, all_papers)
        self.dispatched_field_papers = load_gzip(path)
        keys = list(self.dispatched_field_papers)[idx_start:idx_finish]
        self.dispatched_field_papers = {elt: self.dispatched_field_papers[elt] for elt in keys}
        print(' >> Len papers: ', len(self.dispatched_field_papers))


    def get_papers_ref_fields(self):
        '''
        get papers of reference fields
        :param fields:
        :return:
        '''
        for field in self.ref_fields:
            path_papers = os.path.join(path_produced,field,self.all_papers)
            papers_field = load_gzip(path_papers)
            self.papers_ref_fields.update(papers_field)

        self.papers_ref_fields['fnames'] = [fname for fname in self.papers_ref_fields]


    def paper_in_refs(self,dispatched_paper_name: str) -> bool:
        '''
        check if paper exists in refs
        :param dispatched_paper_name:
        :return:
        '''
        for fname in self.papers_ref_fields:
            if dispatched_paper_name in fname:
                return True
        return False

    def get_in_refs_papers(self,dispatched_paper_name: str) -> dict:
        '''
        check if the same paper was produced in other fields, and return dictionary with omr if the paper exists
        :return:
        '''
        omr = {}
        if not self.papers_ref_fields:
            self.get_papers_ref_fields()

        if self.paper_in_refs(dispatched_paper_name):
            for fname in self.papers_ref_fields:
                if dispatched_paper_name in fname:
                    omr[fname] = self.papers_ref_fields[fname]
        return omr

    def produce_paper(self, paper: dict, paper_name: str) -> dict:
        '''
        produce paper using Paper_Producer : turns paper 'fname' to 'fname_objectives',
        'fname_methods' and 'fname_results'
        :param paper:
        :param paper_name:
        :return:
        '''
        prod = self.get_in_refs_papers(paper_name)
        if prod:
            self.produced_field_papers.update(prod)
        else:
            pap_producer = Paper_Producer(paper=paper, paper_name=paper_name,
                                  obj_classifier=self.obj_classifier,
                                  bomr_classifier=self.bomr_classifier,
                                  nlp=self.nlp)

            pap_producer.produce_paper()
            prod = pap_producer.production_paper

            if prod:
              self.produced_field_papers[paper_name+'_objectives'] = prod['objectives']
              self.produced_field_papers[paper_name+'_methods'] = prod['methods']
              self.produced_field_papers[paper_name+'_results'] = prod['results']

    def txt_to_encode(self,fname: str,papers_dict=None)-> str:
      '''
      get what text to encode from papers_dict (element of self.produced_field_papers)
      :param fname:
      :return:
      '''
      if not papers_dict:
          papers_dict = self.produced_field_papers

      messages = ' '.join(papers_dict[fname]['messages'])
      if '_objectives' in fname:
        return papers_dict[fname]['title']+ ' ' + messages
      else:
        return messages

    def gather_production_papers(self,papers_name='',save=True):
        '''
        load all_papers chunks (produced between 2 indices for same database) and gather them in the same dictionary
        :return: dictionary
        '''
        path_produced_papers = os.path.join(path_produced, self.field)
        all_files = os.listdir(path_produced_papers)

        if papers_name:
            new_fnames = {papers_name: [elt for elt in all_files if papers_name in elt]}
            disp_zone_fnames = [papers_name,]

        else:
            disp_zone_path = os.path.join(path_dispatched, self.field)
            disp_zone_fnames = os.listdir(disp_zone_path)

            all_files = os.listdir(path_produced_papers)

            new_fnames = {fname: [elt for elt in all_files if re.findall(fname + '_\d+', elt)] for fname in disp_zone_fnames}

        for fname in disp_zone_fnames:
            if not new_fnames[fname]:
                new_fnames[fname] = [fname]

        all_papers = {}
        print('>> Loading for updating..')
        for fname in new_fnames:
            fnames = new_fnames[fname]
            path_fnames = [os.path.join(path_produced_papers,fname) for fname in fnames]
            all_papers[fname]= load_gzip_and_update(path_fnames)

        if save:
            print('>> Saving..')
            for fname in all_papers:
                path = os.path.join(path_produced_papers,fname)
                print('  Saving : ', fname)
                save_gzip(path,all_papers[fname])
                for paper in all_papers[fname]:
                    self.produced_field_papers[paper]= all_papers[fname][paper]

    def remove_chunks(self,all_papers_name):
        '''
        remove all_papers chunks (produced between 2 indices for same database)
        :return:
        '''
        path_produced_papers = os.path.join(path_produced, self.field)
        all_files = os.listdir(path_produced_papers)

        chunks_fname = [elt for elt in all_files if re.findall(all_papers_name + '_\d+', elt)]
        paths = [os.path.join(path_produced_papers,elt) for elt in chunks_fname]
        for path in paths:
            os.remove(path)

    def get_field_index(self):
        '''
        get field faiss index by computing the IVFPQ index
        :return:
        '''

        if self.produced_field_papers:
            fnames = list(self.produced_field_papers.keys())
        else:
            self.produced_field_papers,_ = self.load_combine_produced_allpapers()
            fnames = list(self.produced_field_papers.keys())

        print('>> Encoding ..')
        to_encode = [self.txt_to_encode(fn) for fn in fnames]
        encoded_fields = self.encoder.encode(to_encode)
        encoded_fields = np.asarray(encoded_fields.astype('float32'))

        print('>> Getting ids ..')
        # self.field_index = faiss.IndexIVFPQ(vecs,D,ncentroids,code_size ,8)
        # self.field_index.train(encoded_fields)
        # self.field_index.add(encoded_fields)

    def update_field_index(self,new_all_papers: str):
        '''
        update existing field index with new all_papers
        :param new_all_papers:
        :return:
        '''
        print(f'Loading new produced papers : {new_all_papers}')
        prod_zone = Production_Zone()
        papers_dict = prod_zone.get_papers(field=self.field, fname=new_all_papers, verbose=True)
        fnames = list(papers_dict.keys())

        print('>> Encoding ..')
        to_encode = [self.txt_to_encode(fn,papers_dict) for fn in fnames]
        encoded_fields = self.encoder.encode(to_encode)
        encoded_fields = np.asarray(encoded_fields.astype('float32'))

        print('>> Adding ids ..')
        if self.field_index.is_trained:
            print('  >> Index is trained !')
            self.field_index.add(encoded_fields)
        else:
            print('  >> Index needs to be retrained ! ')


    def load_field_index(self):
        '''
        Load field faiss index if it exists
        :return:
        '''
        if not self.field_index:
          path = os.path.join(path_produced,self.field,'encodings.index')
          # self.field_index = faiss.read_index(path)

    def load_combine_produced_allpapers(self, size_data=0):
        '''
        load & combine production papers in same dictionary. If size_data, take percentage of each all_papers db.
        (same code as in Dispatched_Zone, get_field in naimai.pipelines.zones)
        :return:
        '''
        print('>> Loading production papers')
        path_produced_papers = os.path.join(path_produced, self.field)
        all_files = os.listdir(path_produced_papers)
        paths = [os.path.join(path_produced_papers,fl) for fl in all_files]
        paths = [elt for elt in paths if os.path.isfile(elt) and 'encoding' not in elt and 'sqlite' not in elt] # keep only files

        size_each_all_papers = 0
        if size_data:
            size_each_all_papers = int(size_data/len(paths))+1
            print('>> Each all_papers size : ', size_each_all_papers)

        all_paps = {}
        eval_all_paps= {}
        for p in paths:
            data = load_gzip(p)
            if size_each_all_papers:
                keys = list(data.keys())
                if len(data)>size_each_all_papers:
                    keys_selected= random.sample(keys,size_each_all_papers)
                else:
                    keys_selected = keys
                data = {key: data[key] for key in keys_selected}

            all_paps.update(data)
            # eval_all_paps.update(eval_data)
        return all_paps,eval_all_paps

    def fine_tune_field_encoder(self, size_data: int,save_model: bool,batch_size=16,n_epochs=10):
        '''
        load & combine all the "all papers" & finetune search model using a size 'size_data' from field_papers
        :param size_data:
        :param save_model:
        :param batch_size:
        :param n_epochs:
        :return:
        '''
        # combined_papers = self.load_combine_dispatched_allpapers(size_data)
        combined_papers,combined_eval_papers = self.load_combine_produced_allpapers(size_data=size_data)
        print('Len everything : ', len(combined_papers))
        self.smodel = Search_Model(field=self.field, papers=combined_papers,eval_papers=combined_eval_papers,
                                   batch_size=batch_size, n_epochs=n_epochs)
        if self.encoder:
            self.smodel.model = self.encoder
        self.smodel.fine_tune()
        self.encoder = self.smodel.model
        if save_model:
          self.save_model()

    def save_model(self):
        '''
        Saving the encoder model (when trained on the field papers)
        :return:
        '''
        path = os.path.join(path_produced, self.field, 'search_model')
        self.encoder.save(path)

    def produce_field_papers(self,show_tqdm=True):
        '''
        Produce papers contained in field (dispatched files).
        :return:
        '''
        print('>> Producing field papers..')
        if show_tqdm:
            range_ =  tqdm(self.dispatched_field_papers)
        else:
            range_ = self.dispatched_field_papers

        for fname in range_:
            pap = self.dispatched_field_papers[fname]
            if pap['Abstract']:
                try:
                    self.produce_paper(paper=pap, paper_name=fname)
                except:
                    pass
        print(' ')

    # def produce(self,save_papers=False,save_field_index=False):
    #     '''
    #     Produce field papers > getting field faiss index.
    #     :param save_papers:
    #     :param save_field_index:
    #     :return:
    #     '''
    #     if self.encoder:
    #         # produce field papers
    #         self.produce_field_papers()
    #
    #         #compute Faiss index
    #         self.get_field_index()
    #
    #         #save
    #         if save_papers:
    #             print('>> Saving papers..')
    #             self.save_papers()
    #
    #         if save_field_index:
    #             print('>> Saving field index..')
    #             self.save_field_index()
    #         print(' ')
    #         print('>> Done!')
    #     else:
    #         print('>> The production is not done (no field encoder). You need to fine tune.')

    def save_field_index(self):
        '''
        save field faiss index for semantic search
        :return:
        '''
        path = os.path.join(path_produced, self.field, 'encodings.index')
        # faiss.write_index(self.field_index, path)

    def save_papers(self):
        '''
        save produced papers
        :return:
        '''
        if self.extracted:
            file_name = f'{self.all_papers}_{self.idx_start}_{self.idx_finish}'
        else:
            file_name = self.all_papers
        path = os.path.join(path_produced, self.field, file_name)
        save_gzip(path, self.produced_field_papers)


class Custom_Producer:
    '''
    1 Takes formatted custom papers in dictionary all_papers and transform them into a produced all_paper using Paper Producer obj
    2 no need of field index
    '''
    def __init__(self, papers_dict):
        self.obj_classifier = None
        self.bomr_classifier = None
        self.nlp = None
        self.smodel = None
        self.papers_ref_fields = {} # fields used to check if same paper is already produced

        self.all_papers = papers_dict

        self.produced_custom_papers = {}
        self.produce_only_fnames=False

        self.load_obj_classifier()
        self.load_bomr_classifier()
        self.load_nlp()

    def load_obj_classifier(self):
        '''
        Load the objective classifier: model that selects the objective sentence in the abstract.
        :param obj_classifier:
        :return:
        '''
        print('>> Loading obj classifier..')
        self.obj_classifier = Objective_classifier()

    def load_bomr_classifier(self):
        '''
        Load the bomr classifier, the model that structures the abstract into bibliography, objectives, methods
        and results.
        :param bomr_classifier:
        :return:
        '''
        print('>> Loading bomr classifier..')
        self.bomr_classifier = NER_BOMR_classifier(load_model=True, path_model=path_bomr_classifier,
                                                       predict_mode=True)
    def load_nlp(self):
        '''
        Load nlp model (spaCy)
        :param nlp:
        :return:
        '''
        print('>> Loading nlp..')
        self.nlp = spacy.load(nlp_vocab)

    def produce_paper(self, paper: dict, paper_name: str) -> dict:
        '''
        produce paper using Paper_Producer : turns paper 'fname' to 'fname_objectives',
        'fname_methods' and 'fname_results'
        :param paper:
        :param paper_name:
        :return:
        '''
        pap_producer = Paper_Producer(paper=paper, paper_name=paper_name,
                              obj_classifier=self.obj_classifier,
                              bomr_classifier=self.bomr_classifier,
                              is_custom=True,
                              nlp=self.nlp)

        pap_producer.produce_paper()
        prod = pap_producer.production_paper

        if prod:
          self.produced_custom_papers[paper_name+'_objectives'] = prod['objectives']
          self.produced_custom_papers[paper_name+'_methods'] = prod['methods']
          self.produced_custom_papers[paper_name+'_results'] = prod['results']

    def txt_to_encode(self,fname: str,papers_dict=None)-> str:
      '''
      get what text to encode from papers_dict (element of self.produced_custom_papers)
      :param fname:
      :return:
      '''
      if not papers_dict:
          papers_dict = self.produced_custom_papers

      messages = ' '.join(papers_dict[fname]['messages'])
      if '_objectives' in fname:
        return papers_dict[fname]['title']+ ' ' + messages
      else:
        return messages

    def produce_custom_papers(self,show_tqdm=True):
        '''
        produce read papers.
        :return:
        '''
        print('>> Producing custom papers..')
        if show_tqdm:
            range_ = tqdm(self.all_papers)
        else:
            range_ = self.all_papers
        for fname in range_:
            pap = self.all_papers[fname]
            if pap['Abstract']:
                # try:
                self.produce_paper(paper=pap, paper_name=fname)
                # except:
                #     pass
        print(' ')


import random
from keybert import KeyBERT


class QueryGeneration:
    def __init__(self, nb_queries=2):
        self.nb_queries = nb_queries
        # self.queries = []
        self.keybert = KeyBERT()

    def from_message(self, message: str) -> list:
        '''
        get 2 keywords from message
        :param message:
        :return:
        '''
        keywords = self.keybert.extract_keywords(message, keyphrase_ngram_range=(2, 5), top_n=self.nb_queries)
        queries = [elt[0] for elt in keywords]
        return queries

    # def from_produced_paper(self,paper):
    #     if 'title' in paper.keys():
    #         title = paper['title']
    #         self.queries += self.from_message(title)

    #     messages = paper['messages']

    
    # def from_paragraphs(self, nb_queries_to_add):
    #     paragraph = self.paper['Abstract'].strip()+ ' ' + self.paper['Title'].strip()
    #     if paragraph:
    #         keywords = self.keybert.extract_keywords(paragraph, keyphrase_ngram_range=(1, 3),top_n=nb_queries_to_add)
    #         queries = [elt[0] for elt in keywords]
    #         try:
    #             three_random = random.sample(queries, k=nb_queries_to_add)
    #             self.queries += three_random
    #         except:
    #             pass
    #             # print('problem in paper {}, data base {}'.format(self.paper['file_name'],self.paper['database']))


    # def generate(self):
    #     self.queries = []
    #     self.from_keywords()
    #     self.from_title()
    #     self.queries = [elt for elt in self.queries if elt]
    #     len_queries = len(self.queries)
    #     nb_queries_to_add = self.nb_queries - len_queries
    #     self.from_paragraphs(nb_queries_to_add)



from unittest import TestCase
from naimai.papers.full_text.pdf_paper import papers_pdf
from naimai.pipelines.producers import Custom_Producer
import os

class CustomProducerTest(TestCase):
    '''
    Test custom producer, by producing pdf papers.
    '''
    @classmethod
    def setUpClass(cls) -> None:
        cls.fnames = ['Bousmar1999.pdf_objectives', 'Bousmar1999.pdf_methods', 'Bousmar1999.pdf_results', 'Chen2016.pdf_objectives', 'Chen2016.pdf_methods', 'Chen2016.pdf_results']
        cls.fname_test = 'Chen2016.pdf_objectives'
        cls.title = 'Determination of Apparent Shear Stress and its Application in Compound Channels'
        cls.numcitedby = 10
        cls.journal = 'Procedia Engineering'
        cls.authors = 'Chen et al. (2016)'
        cls.first_message = 'The momentum exchange in the mixing zone of compound channel is a common phenomenon and has an important effect on discharge estimation.'
        cls.reported = 'Chen et al. (2016) quantified with apparent shear stress (ASS) at the interface of floodplain and main channel.'



        path = os.path.join('tests', 'papers', 'input_data')
        obj = papers_pdf(papers_path=path)
        obj.get_papers()

        producer = Custom_Producer(papers_dict=obj.elements)
        producer.produce_custom_papers(show_tqdm=False)

        result = producer.produced_custom_papers
        cls.keys = list(result.keys())
        cls.first_paper = result[cls.keys[3]]

    def test_fnames(self):
        '''
        test if correct fnames are read
        :return:
        '''
        self.assertEqual(self.keys, self.fnames)

    def test_title(self):
        '''
        test title of the first paper
        :return:
        '''
        title_first_paper = self.first_paper['title']
        self.assertEqual(title_first_paper, self.title)

    def test_numcitedby(self):
        '''
        test numcitedby of the first paper
        :return:
        '''
        numcitedby_first_paper = self.first_paper['numCitedBy']
        self.assertEqual(numcitedby_first_paper, self.numcitedby)

    def test_journal(self):
        '''
        test journal of the first paper
        :return:
        '''
        journal_first_paper = self.first_paper['journal']
        self.assertEqual(journal_first_paper, self.journal)

    def test_authors(self):
        '''
        test authors of the first paper
        :return:
        '''
        authors_first_paper = self.first_paper['authors']
        self.assertEqual(authors_first_paper, self.authors)

    def test_first_msg(self):
        '''
        test first message of the first paper
        :return:
        '''
        message_first_paper = self.first_paper['messages'][0]
        self.assertEqual(message_first_paper, self.first_message)

    def test_reported(self):
        '''
        test first message of the first paper
        :return:
        '''
        reported_first_paper = self.first_paper['reported']
        self.assertEqual(reported_first_paper, self.reported)


output_labels = ['O', 'B-background', 'I-background', 'B-objectives', 'I-objectives',
                         'B-methods', 'I-methods', 'B-results', 'I-results']
ner_labels = ['background', 'objectives', 'methods','results', "other"]
ner_labels2 = ['BAC', 'OBJ', 'MET','RES', "OTH"]
list_colors = ['#2adddd',"#80ffb4","#ff8042","#2b7ff6","#007f00"]
colors_labels = {key: color for key,color in zip(ner_labels2,list_colors)}
threshold_tf_similarity = .05